package config

import (
	"os"
	"time"

	"github.com/google/uuid"
	"github.com/sirupsen/logrus"
	log "github.com/sirupsen/logrus"
	"github.com/spf13/viper"
)

const (
	defaultTimeOutDuration10minutes = time.Minute * 10
	basePNDUUIDKey                  = "basePNDUUID"
	baseSouthBoundTypeKey           = "baseSouthBoundType"
	baseSouthBoundUUIDKey           = "baseSouthBoundUUID"
	changeTimeoutKey                = "GOSDN_CHANGE_TIMEOUT"
)

// BasePndUUID is an uuid for the base PND
var BasePndUUID uuid.UUID

// BaseSouthBoundType is the type of the base SBI
var BaseSouthBoundType int32

// BaseSouthBoundUUID is an uuid for the base SBI
var BaseSouthBoundUUID uuid.UUID

// ChangeTimeout is the default timeout for a change
var ChangeTimeout time.Duration

// LogLevel ist the default log level
var LogLevel logrus.Level

// Init gets called on module import
func Init() {
	InitializeConfig()
}

// InitializeConfig loads the configuration
func InitializeConfig() error {
	var err error

	basePNDUUIDFromViper, err := getUUIDFromViper(basePNDUUIDKey)
	if err != nil {
		return err
	}

	BasePndUUID = basePNDUUIDFromViper

	baseSouthBoundUUIDFromViper, err := getUUIDFromViper(baseSouthBoundUUIDKey)
	if err != nil {
		return err
	}

	BaseSouthBoundUUID = baseSouthBoundUUIDFromViper

	BaseSouthBoundType = viper.GetInt32(baseSouthBoundTypeKey)
	if BaseSouthBoundType != 0 {
		viper.Set(baseSouthBoundTypeKey, 0)
		viper.WriteConfig()
	}

	err = setChangeTimeout()
	if err != nil {
		return err
	}

	setLogLevel()

	return nil
}

func getUUIDFromViper(viperKey string) (uuid.UUID, error) {
	UUIDAsString := viper.GetString(viperKey)
	if UUIDAsString == "" {
		newUUID := uuid.New()
		viper.Set(viperKey, newUUID.String())
		viper.WriteConfig()

		return newUUID, nil
	}

	parsedUUID, err := uuid.Parse(UUIDAsString)
	if err != nil {
		return uuid.Nil, err
	}

	return parsedUUID, nil
}

func setChangeTimeout() error {
	e := os.Getenv(changeTimeoutKey)
	if e != "" {
		changeTimeout, err := time.ParseDuration(e)
		if err != nil {
			log.Fatal(err)
		}
		ChangeTimeout = changeTimeout
	} else {
		ChangeTimeout = time.Minute * 10
	}

	return nil
}

func setLogLevel() {
	if os.Getenv("GOSDN_LOG") == "nolog" {
		LogLevel = logrus.PanicLevel
	} else {
		LogLevel = logrus.InfoLevel
	}
}
