package server

import (
	"context"
	"time"

	pb "code.fbi.h-da.de/danet/api/go/gosdn/core"
	ppb "code.fbi.h-da.de/danet/api/go/gosdn/pnd"
	spb "code.fbi.h-da.de/danet/api/go/gosdn/southbound"
	"code.fbi.h-da.de/danet/gosdn/metrics"
	"code.fbi.h-da.de/danet/gosdn/nucleus"
	"github.com/google/uuid"
	"github.com/prometheus/client_golang/prometheus"
)

type core struct {
	pb.UnimplementedCoreServer
}

func (s core) Get(ctx context.Context, request *pb.GetRequest) (*pb.GetResponse, error) {
	labels := prometheus.Labels{"service": "core", "rpc": "get"}
	start := metrics.StartHook(labels, grpcRequestsTotal)
	defer metrics.FinishHook(labels, start, grpcRequestDurationSecondsTotal, grpcRequestDurationSeconds)
	var pndList []uuid.UUID
	switch request.All {
	case true:
		pndList = pndc.UUIDs()
	default:
		var err error
		pndList, err = stringToUUID(request.Pid)
		if err != nil {
			return nil, handleRPCError(labels, err)
		}
	}

	pnds := make([]*ppb.PrincipalNetworkDomain, len(pndList))
	for i, id := range pndList {
		pnd, err := pndc.GetPND(id)
		if err != nil {
			return nil, err
		}

		ond, err := fillOnds(pnd, true)
		if err != nil {
			return nil, handleRPCError(labels, err)
		}

		sbi, err := fillSbis(pnd, true)
		if err != nil {
			return nil, handleRPCError(labels, err)
		}

		pnds[i] = &ppb.PrincipalNetworkDomain{
			Id:          pnd.ID().String(),
			Name:        pnd.GetName(),
			Description: pnd.GetDescription(),
			Ond:         ond,
			Sbi:         sbi,
		}
	}
	return &pb.GetResponse{
		Timestamp: time.Now().UnixNano(),
		Pnd:       pnds,
	}, nil
}

func (s core) Set(ctx context.Context, request *pb.SetRequest) (*pb.SetResponse, error) {
	labels := prometheus.Labels{"service": "core", "rpc": "set"}
	start := metrics.StartHook(labels, grpcRequestsTotal)
	defer metrics.FinishHook(labels, start, grpcRequestDurationSecondsTotal, grpcRequestDurationSeconds)
	for _, r := range request.Pnd {
		sbi := nucleus.NewSBI(spb.Type_OPENCONFIG)

		pnd, err := nucleus.NewPND(r.Name, r.Description, uuid.New(), sbi, nil, nil)
		if err != nil {
			return nil, handleRPCError(labels, err)
		}
		if err := pndc.Add(pnd); err != nil {
			return nil, handleRPCError(labels, err)
		}
	}
	return &pb.SetResponse{
		Timestamp: time.Now().UnixNano(),
		Status:    pb.SetResponse_OK,
	}, nil
}
