package server

import (
	"context"
	"reflect"
	"time"

	"google.golang.org/grpc/codes"
	"google.golang.org/grpc/status"

	ppb "code.fbi.h-da.de/danet/api/go/gosdn/pnd"
	spb "code.fbi.h-da.de/danet/api/go/gosdn/southbound"
	"code.fbi.h-da.de/danet/gosdn/interfaces/networkdomain"
	"code.fbi.h-da.de/danet/gosdn/metrics"
	"code.fbi.h-da.de/danet/gosdn/nucleus/errors"
	"github.com/google/uuid"
	"github.com/openconfig/ygot/ygot"
	"github.com/prometheus/client_golang/prometheus"
	log "github.com/sirupsen/logrus"
)

type pndServer struct {
	ppb.UnimplementedPndServer
}

func (p pndServer) Get(ctx context.Context, request *ppb.GetRequest) (*ppb.GetResponse, error) {
	labels := prometheus.Labels{"service": "pnd", "rpc": "get"}
	start := metrics.StartHook(labels, grpcRequestsTotal)
	defer metrics.FinishHook(labels, start, grpcRequestDurationSecondsTotal, grpcRequestDurationSeconds)
	pid, err := uuid.Parse(request.Pid)
	if err != nil {
		return nil, handleRPCError(labels, err)
	}
	switch req := request.Request.(type) {
	case *ppb.GetRequest_Pnd:
		return handleGetPnd(pid)
	case *ppb.GetRequest_Ond:
		return handleGetOnd(pid, req)
	case *ppb.GetRequest_Sbi:
		return handleGetSbi(pid, req)
	case *ppb.GetRequest_Change:
		return handleGetChange(pid, req)
	case *ppb.GetRequest_Path:
		return handleGetPath(pid, req)
	default:
		return nil, handleRPCError(labels, errors.ErrOperationNotSupported{Op: reflect.TypeOf(request.Request)})
	}
}

func handleGetPnd(pid uuid.UUID) (*ppb.GetResponse, error) {
	pnd, err := pndc.GetPND(pid)
	if err != nil {
		log.Error(err)
		return nil, status.Errorf(codes.Aborted, "%v", err)
	}
	onds, err := fillOnds(pnd, true)
	if err != nil {
		log.Error(err)
		return nil, status.Errorf(codes.Aborted, "%v", err)
	}
	sbis, err := fillSbis(pnd, true)
	if err != nil {
		log.Error(err)
		return nil, status.Errorf(codes.Aborted, "%v", err)
	}
	changes, err := fillChanges(pnd, true)
	if err != nil {
		log.Error(err)
		return nil, status.Errorf(codes.Aborted, "%v", err)
	}

	return &ppb.GetResponse{
		Timestamp: time.Now().UnixNano(),
		Pnd: &ppb.PrincipalNetworkDomain{
			Id:          pnd.ID().String(),
			Name:        pnd.GetName(),
			Description: pnd.GetDescription(),
			Ond:         onds,
			Sbi:         sbis,
			Change:      changes,
		},
	}, nil
}

func handleGetSbi(pid uuid.UUID, req *ppb.GetRequest_Sbi) (*ppb.GetResponse, error) {
	pnd, err := pndc.GetPND(pid)
	if err != nil {
		log.Error(err)
		return nil, status.Errorf(codes.Aborted, "%v", err)
	}
	sbis, err := fillSbis(pnd, req.Sbi.GetAll(), req.Sbi.Sid...)
	if err != nil {
		log.Error(err)
		return nil, status.Errorf(codes.Aborted, "%v", err)
	}
	return &ppb.GetResponse{
		Timestamp: time.Now().UnixNano(),
		Sbi:       sbis,
	}, nil
}

func stringToUUID(sid []string) ([]uuid.UUID, error) {
	UUIDs := make([]uuid.UUID, len(sid))
	for i, id := range sid {
		parsed, err := uuid.Parse(id)
		if err != nil {
			log.Error(err)
			return nil, status.Errorf(codes.Aborted, "%v", err)
		}
		UUIDs[i] = parsed
	}
	return UUIDs, nil
}

func handleGetOnd(pid uuid.UUID, req *ppb.GetRequest_Ond) (*ppb.GetResponse, error) {
	pnd, err := pndc.GetPND(pid)
	if err != nil {
		log.Error(err)
		return nil, status.Errorf(codes.Aborted, "%v", err)
	}
	onds, err := fillOnds(pnd, req.Ond.All, req.Ond.Did...)
	if err != nil {
		log.Error(err)
		return nil, status.Errorf(codes.Aborted, "%v", err)
	}
	return &ppb.GetResponse{
		Timestamp: time.Now().UnixNano(),
		Ond:       onds,
	}, nil
}

func handleGetChange(pid uuid.UUID, req *ppb.GetRequest_Change) (*ppb.GetResponse, error) {
	pnd, err := pndc.GetPND(pid)
	if err != nil {
		log.Error(err)
		return nil, status.Errorf(codes.Aborted, "%v", err)
	}
	changes, err := fillChanges(pnd, req.Change.All, req.Change.Cuid...)
	if err != nil {
		log.Error(err)
		return nil, status.Errorf(codes.Aborted, "%v", err)
	}
	return &ppb.GetResponse{
		Timestamp: time.Now().UnixNano(),
		Change:    changes,
	}, nil
}

func handleGetPath(pid uuid.UUID, req *ppb.GetRequest_Path) (*ppb.GetResponse, error) {
	pnd, err := pndc.GetPND(pid)
	if err != nil {
		log.Error(err)
		return nil, status.Errorf(codes.Aborted, "%v", err)
	}
	duid, err := uuid.Parse(req.Path.Did)
	if err != nil {
		log.Error(err)
		return nil, status.Errorf(codes.Aborted, "%v", err)
	}
	_, err = pnd.Request(duid, req.Path.Path)
	if err != nil {
		log.Error(err)
		return nil, status.Errorf(codes.Aborted, "%v", err)
	}
	ond, err := fillOnds(pnd, false, req.Path.Did)
	if err != nil {
		log.Error(err)
		return nil, status.Errorf(codes.Aborted, "%v", err)
	}
	return &ppb.GetResponse{
		Timestamp: time.Now().UnixNano(),
		Ond:       ond,
	}, nil
}

func fillSbis(pnd networkdomain.NetworkDomain, all bool, sid ...string) ([]*spb.SouthboundInterface, error) {
	var sbiList []uuid.UUID

	sbiStore := pnd.GetSBIs()
	switch all {
	case true:
		sbiList = sbiStore.UUIDs()
	default:
		var err error
		if len(sid) == 0 {
			return nil, &errors.ErrInvalidParameters{
				Func:  fillSbis,
				Param: "length of 'sid' cannot be '0' when 'all' is set to 'false'",
			}
		}
		sbiList, err = stringToUUID(sid)
		if err != nil {
			log.Error(err)
			return nil, status.Errorf(codes.Aborted, "%v", err)
		}
	}
	sbis := make([]*spb.SouthboundInterface, len(sbiList))
	for i, id := range sbiList {
		_, err := sbiStore.Get(id)
		if err != nil {
			log.Error(err)
			return nil, status.Errorf(codes.Aborted, "%v", err)
		}
		sbis[i] = &spb.SouthboundInterface{
			Id: id.String(),
		}
	}
	return sbis, nil
}

func fillOnds(pnd networkdomain.NetworkDomain, all bool, did ...string) ([]*ppb.OrchestratedNetworkingDevice, error) {
	var ondList []uuid.UUID
	var onds []*ppb.OrchestratedNetworkingDevice

	// all indicates if a client wants all devices or only a single one
	switch all {
	case true:
		ondList = pnd.Devices()
		onds = make([]*ppb.OrchestratedNetworkingDevice, len(ondList))
		for _, id := range ondList {
			did = append(did, id.String())
		}
	default:
		if len(did) == 0 {
			err := &errors.ErrInvalidParameters{
				Func:  fillOnds,
				Param: "length of 'did' cannot be '0' when 'all' is set to 'false'",
			}
			log.Error(err)

			return nil, err
		}

		onds = make([]*ppb.OrchestratedNetworkingDevice, 1)
	}

	for i, id := range did {
		d, err := pnd.GetDevice(id)
		if err != nil {
			log.Error(err)
			return nil, status.Errorf(codes.Aborted, "%v", err)
		}
		cfg := ygot.GNMINotificationsConfig{}
		dev, err := ygot.TogNMINotifications(d.Model(), time.Now().UnixNano(), cfg)
		if err != nil {
			log.Error(err)
			return nil, status.Errorf(codes.Aborted, "%v", err)
		}
		onds[i] = &ppb.OrchestratedNetworkingDevice{
			Id:     id,
			Name:   d.Name(),
			Device: dev,
		}
	}

	return onds, nil
}

func fillChanges(pnd networkdomain.NetworkDomain, all bool, cuid ...string) ([]*ppb.Change, error) {
	var changeList []uuid.UUID

	switch all {
	case true:
		changeList = pnd.PendingChanges()
		changeList = append(changeList, pnd.CommittedChanges()...)
	default:
		var err error
		if len(cuid) == 0 {
			return nil, &errors.ErrInvalidParameters{
				Func:  fillOnds,
				Param: "length of 'did' cannot be '0' when 'all' is set to 'false'",
			}
		}
		changeList, err = stringToUUID(cuid)
		if err != nil {
			log.Error(err)
			return nil, status.Errorf(codes.Aborted, "%v", err)
		}
	}

	changes := make([]*ppb.Change, len(changeList))
	for i, ch := range changeList {
		c, err := pnd.GetChange(ch)
		if err != nil {
			log.Error(err)
			return nil, status.Errorf(codes.Aborted, "%v", err)
		}

		changes[i] = &ppb.Change{
			Id:    ch.String(),
			Age:   c.Age().Microseconds(),
			State: c.State(),
		}
	}
	return changes, nil
}

func (p pndServer) Set(ctx context.Context, request *ppb.SetRequest) (*ppb.SetResponse, error) {
	labels := prometheus.Labels{"service": "pnd", "rpc": "set"}
	start := metrics.StartHook(labels, grpcRequestsTotal)
	defer metrics.FinishHook(labels, start, grpcRequestDurationSecondsTotal, grpcRequestDurationSeconds)
	pid, err := uuid.Parse(request.Pid)
	if err != nil {
		return nil, handleRPCError(labels, err)
	}

	pnd, err := pndc.GetPND(pid)
	if err != nil {
		return nil, handleRPCError(labels, err)
	}

	ondResp, err := handleSetOnd(pnd, request.Ond)
	if err != nil {
		return nil, handleRPCError(labels, err)
	}
	changeResp, err := handleSetChange(pnd, request.Change)
	if err != nil {
		return nil, handleRPCError(labels, err)
	}
	changeRequestResp, err := handleChangeRequest(pnd, request.ChangeRequest)
	if err != nil {
		return nil, handleRPCError(labels, err)
	}
	return &ppb.SetResponse{
		Timestamp: time.Now().UnixNano(),
		Status:    ppb.SetResponse_OK,
		Responses: []*ppb.SetResponse{
			ondResp,
			changeResp,
			changeRequestResp,
		},
	}, nil
}

func handleSetOnd(pnd networkdomain.NetworkDomain, req []*ppb.SetOnd) (*ppb.SetResponse, error) {
	for _, r := range req {
		sid, err := uuid.Parse(r.Sbi.Id)
		if err != nil {
			log.Error(err)
			return nil, status.Errorf(codes.Aborted, "%v", err)
		}
		if err := pnd.AddDevice(r.DeviceName, r.TransportOption, sid); err != nil {
			log.Error(err)
			return nil, status.Errorf(codes.Aborted, "%v", err)
		}
	}
	return &ppb.SetResponse{
		Timestamp: time.Now().UnixNano(),
		Status:    ppb.SetResponse_OK,
	}, nil
}

func handleSetChange(pnd networkdomain.NetworkDomain, req []*ppb.SetChange) (*ppb.SetResponse, error) {
	for _, r := range req {
		cuid, err := uuid.Parse(r.Cuid)
		if err != nil {
			log.Error(err)
			return nil, status.Errorf(codes.Aborted, "%v", err)
		}
		switch r.Op {
		case ppb.SetChange_COMMIT:
			if err := pnd.Commit(cuid); err != nil {
				log.Error(err)
				return nil, status.Errorf(codes.Aborted, "%v", err)
			}
		case ppb.SetChange_CONFIRM:
			if err := pnd.Confirm(cuid); err != nil {
				log.Error(err)
				return nil, status.Errorf(codes.Aborted, "%v", err)
			}
		default:
			return nil, &errors.ErrInvalidParameters{
				Func:  handleSetChange,
				Param: r.Op,
			}
		}
	}
	return &ppb.SetResponse{
		Timestamp: time.Now().UnixNano(),
		Status:    ppb.SetResponse_OK,
	}, nil
}

func handleChangeRequest(pnd networkdomain.NetworkDomain, req []*ppb.ChangeRequest) (*ppb.SetResponse, error) {
	for _, r := range req {
		did, err := uuid.Parse(r.Id)
		if err != nil {
			log.Error(err)
			return nil, status.Errorf(codes.Aborted, "%v", err)
		}
		// TODO: Return CUID in API
		_, err = pnd.ChangeOND(did, r.ApiOp, r.Path, r.Value)
		if err != nil {
			log.Error(err)
			return nil, status.Errorf(codes.Aborted, "%v", err)
		}
	}
	return &ppb.SetResponse{
		Timestamp: time.Now().UnixNano(),
		Status:    ppb.SetResponse_OK,
	}, nil
}

func (p pndServer) Delete(ctx context.Context, req *ppb.DeleteRequest) (*ppb.DeleteResponse, error) {
	pid, err := uuid.Parse(req.Pid)
	if err != nil {
		log.Error(err)
		return nil, status.Errorf(codes.Aborted, "%v", err)
	}
	pnd, err := pndc.GetPND(pid)
	if err != nil {
		log.Error(err)
		return nil, status.Errorf(codes.Aborted, "%v", err)
	}

	did, err := uuid.Parse(req.Uuid)
	if err != nil {
		log.Error(err)
		return nil, status.Errorf(codes.Aborted, "%v", err)
	}
	if err := pnd.RemoveDevice(did); err != nil {
		log.Error(err)
		return nil, status.Errorf(codes.Aborted, "%v", err)
	}
	return &ppb.DeleteResponse{
		Timestamp: time.Now().UnixNano(),
		Status:    ppb.DeleteResponse_OK,
	}, nil
}
