package store

import (
	"log"
	"reflect"
	"testing"

	"code.fbi.h-da.de/danet/api/go/gosdn/pnd"
	ppb "code.fbi.h-da.de/danet/api/go/gosdn/pnd"
	"code.fbi.h-da.de/danet/gosdn/mocks"
	"github.com/google/uuid"
)

func TestChangeStore_Pending(t *testing.T) {
	changeMock := &mocks.Change{}
	changeMock.On("ID").Return(cuid)
	changeMock.On("State").Return(pnd.Change_PENDING)

	store := NewChangeStore()
	pending := changeMock
	if err := store.Add(pending); err != nil {
		t.Error(err)
		return
	}
	tests := []struct {
		name string
		want []uuid.UUID
	}{
		{
			name: "default",
			want: []uuid.UUID{cuid},
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			s := store
			if got := s.Pending(); !reflect.DeepEqual(got, tt.want) {
				t.Errorf("Pending() = %v, want %v", got, tt.want)
			}
		})
	}
}

func TestChangeStore_Committed(t *testing.T) {
	changeMock := &mocks.Change{}
	changeMock.On("ID").Return(cuid)
	changeMock.On("State").Return(pnd.Change_COMMITTED)
	changeMock.On("Commit").Return(nil)

	store := NewChangeStore()
	committed := changeMock
	if err := committed.Commit(); err != nil {
		t.Error(err)
		return
	}
	if err := store.Add(committed); err != nil {
		t.Error(err)
		return
	}
	tests := []struct {
		name string
		want []uuid.UUID
	}{
		{
			name: "default",
			want: []uuid.UUID{cuid},
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			s := store
			if got := s.Committed(); !reflect.DeepEqual(got, tt.want) {
				t.Errorf("Committed() = %v, want %v", got, tt.want)
			}
		})
	}
}

func TestChangeStore_Confirmed(t *testing.T) {
	changeMock := &mocks.Change{}
	changeMock.On("ID").Return(cuid)
	changeMock.On("State").Return(pnd.Change_CONFIRMED)
	changeMock.On("Commit").Return(nil)
	changeMock.On("Confirm").Return(nil)

	store := NewChangeStore()
	confirmed := changeMock
	if err := confirmed.Commit(); err != nil {
		t.Error(err)
		return
	}
	if err := confirmed.Confirm(); err != nil {
		t.Error(err)
		return
	}
	if err := store.Add(confirmed); err != nil {
		t.Error(err)
		return
	}

	tests := []struct {
		name string
		want []uuid.UUID
	}{
		{
			name: "default",
			want: []uuid.UUID{cuid},
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			s := store
			if got := s.Confirmed(); !reflect.DeepEqual(got, tt.want) {
				t.Errorf("Confirmed() = %v, want %v", got, tt.want)
			}
		})
	}
}

func Test_filterChanges(t *testing.T) {
	var pendingCUID uuid.UUID
	var committedCUID uuid.UUID
	var confirmedCUID uuid.UUID
	var err error

	pendingCUID, err = uuid.Parse("3e8219b0-e926-400d-8660-217f2a25a7c7")
	if err != nil {
		log.Fatal(err)
	}
	committedCUID, err = uuid.Parse("3e8219b0-e926-400d-8660-217f2a25a7c8")
	if err != nil {
		log.Fatal(err)
	}
	confirmedCUID, err = uuid.Parse("3e8219b0-e926-400d-8660-217f2a25a7c9")
	if err != nil {
		log.Fatal(err)
	}

	changeMockPending := &mocks.Change{}
	changeMockPending.On("ID").Return(pendingCUID)
	changeMockPending.On("State").Return(pnd.Change_PENDING)
	changeMockPending.On("Commit").Return(nil)
	changeMockPending.On("Confirm").Return(nil)

	changeMockCommited := &mocks.Change{}
	changeMockCommited.On("ID").Return(committedCUID)
	changeMockCommited.On("State").Return(pnd.Change_COMMITTED)
	changeMockCommited.On("Commit").Return(nil)
	changeMockCommited.On("Confirm").Return(nil)

	changeMockConfirmed := &mocks.Change{}
	changeMockConfirmed.On("ID").Return(confirmedCUID)
	changeMockConfirmed.On("State").Return(pnd.Change_CONFIRMED)
	changeMockConfirmed.On("Commit").Return(nil)
	changeMockConfirmed.On("Confirm").Return(nil)

	store := NewChangeStore()
	pending := changeMockPending
	committed := changeMockCommited
	if err := committed.Commit(); err != nil {
		t.Error(err)
		return
	}
	confirmed := changeMockConfirmed
	if err := confirmed.Commit(); err != nil {
		t.Error(err)
		return
	}
	if err := confirmed.Confirm(); err != nil {
		t.Error(err)
		return
	}
	if err := store.Add(pending); err != nil {
		t.Error(err)
		return
	}
	if err := store.Add(committed); err != nil {
		t.Error(err)
		return
	}
	if err := store.Add(confirmed); err != nil {
		t.Error(err)
		return
	}
	type args struct {
		store *ChangeStore
		state ppb.Change_State
	}
	tests := []struct {
		name string
		args args
		want []uuid.UUID
	}{
		{
			name: "pending",
			args: args{
				store: store,
				state: ppb.Change_PENDING,
			},
			want: []uuid.UUID{pendingCUID},
		},
		{
			name: "committed",
			args: args{
				store: store,
				state: ppb.Change_COMMITTED,
			},
			want: []uuid.UUID{committedCUID},
		},
		{
			name: "confirmed",
			args: args{
				store: store,
				state: ppb.Change_CONFIRMED,
			},
			want: []uuid.UUID{confirmedCUID},
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if got := filterChanges(tt.args.store, tt.args.state); !reflect.DeepEqual(got, tt.want) {
				t.Errorf("filterChanges() = %v, want %v", got, tt.want)
			}
		})
	}
}
