package store

import (
	"reflect"
	"sort"
	"testing"

	"code.fbi.h-da.de/danet/gosdn/interfaces/store"
	"code.fbi.h-da.de/danet/gosdn/mocks"
	"github.com/google/uuid"
)

func Test_Store_add(t *testing.T) {
	type args struct {
		item store.Storable
	}
	tests := []struct {
		name    string
		s       *genericStore
		args    args
		wantErr bool
	}{
		{
			name: "default",
			s:    newGenericStore(),
			args: args{
				item: &mocks.Storable{},
			},
		},
		{
			name: "already exists",
			s: &genericStore{
				Store: map[uuid.UUID]store.Storable{
					iid: &mocks.Storable{},
				},
			},
			args: args{
				item: &mocks.Storable{},
			},
			wantErr: true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			tt.args.item.(*mocks.Storable).On("ID").Return(iid)
			switch tt.name {
			case "already exixts":
				_ = tt.s.Add(tt.args.item)
			default:
			}
			if err := tt.s.Add(tt.args.item); (err != nil) != tt.wantErr {
				t.Errorf("Add() error = %v, wantErr %v", err, tt.wantErr)
			}
		})
	}
}

func Test_Store_delete(t *testing.T) {
	type args struct {
		id uuid.UUID
	}
	tests := []struct {
		name    string
		s       *genericStore
		args    args
		wantErr bool
	}{
		{
			name: "default",
			s: &genericStore{
				Store: map[uuid.UUID]store.Storable{
					iid: &mocks.Storable{},
				},
			},
			args:    args{id: iid},
			wantErr: false,
		},
		{
			name:    "not found empty",
			s:       newGenericStore(),
			args:    args{id: iid},
			wantErr: true,
		},
		{
			name: "not found",
			s: &genericStore{
				Store: map[uuid.UUID]store.Storable{
					iid: &mocks.Storable{},
				},
			},
			args:    args{id: altIid},
			wantErr: true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if err := tt.s.Delete(tt.args.id); (err != nil) != tt.wantErr {
				t.Errorf("delete() error = %v, wantErr %v", err, tt.wantErr)
			}
			if tt.name == "default" {
				item, ok := tt.s.Store[iid]
				if ok {
					t.Errorf("delete() item %v still in genericStore %v", item, tt.s)
				}
			}
		})
	}
}

func Test_Store_exists(t *testing.T) {
	type args struct {
		id uuid.UUID
	}
	tests := []struct {
		name string
		s    *genericStore
		args args
		want bool
	}{
		{
			name: "default",
			s: &genericStore{
				Store: map[uuid.UUID]store.Storable{
					iid: &mocks.Storable{},
				},
			},
			args: args{id: iid},
			want: true,
		},
		{
			name: "not found empty",
			s:    newGenericStore(),
			args: args{id: iid},
			want: false,
		},
		{
			name: "not found",
			s: &genericStore{
				Store: map[uuid.UUID]store.Storable{
					iid: &mocks.Storable{},
				},
			},
			args: args{id: altIid},
			want: false,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if got := tt.s.Exists(tt.args.id); got != tt.want {
				t.Errorf("exists() = %v, want %v", got, tt.want)
			}
		})
	}
}

func Test_Store_get(t *testing.T) {
	type args struct {
		id uuid.UUID
	}
	tests := []struct {
		name    string
		s       *genericStore
		args    args
		want    store.Storable
		wantErr bool
	}{
		{
			name: "exists",
			s: &genericStore{
				Store: map[uuid.UUID]store.Storable{
					iid: &mocks.Storable{},
				},
			},
			args:    args{id: iid},
			want:    &mocks.Storable{},
			wantErr: false,
		},
		{
			name: "not found",
			s: &genericStore{
				Store: map[uuid.UUID]store.Storable{
					iid: &mocks.Storable{},
				},
			},
			args:    args{id: altIid},
			want:    nil,
			wantErr: true,
		},
		{
			name:    "not found empty",
			s:       newGenericStore(),
			args:    args{id: iid},
			want:    nil,
			wantErr: true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			got, err := tt.s.Get(tt.args.id)
			if (err != nil) != tt.wantErr {
				t.Errorf("get() error = %v, wantErr %v", err, tt.wantErr)
				return
			}
			if !reflect.DeepEqual(got, tt.want) {
				t.Errorf("get() got = %v, want %v", got, tt.want)
			}
		})
	}
}

func Test_Store_UUIDs(t *testing.T) {
	tests := []struct {
		name string
		s    *genericStore
		want []uuid.UUID
	}{
		{
			name: "default",
			s: &genericStore{
				Store: map[uuid.UUID]store.Storable{
					iid:    &mocks.Storable{},
					altIid: &mocks.Storable{},
				},
			},
			want: []uuid.UUID{iid, altIid},
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			sort.Slice(tt.want, func(i, j int) bool {
				return tt.want[i].String() < tt.want[j].String()
			})
			got := tt.s.UUIDs()
			sort.Slice(got, func(i, j int) bool {
				return got[i].String() < got[j].String()
			})
			if !reflect.DeepEqual(got, tt.want) {
				t.Errorf("UUIDs() = %v, want %v", got, tt.want)
			}
		})
	}
}
